from .main import TestLineLen
from .main import TestBadWords
from .main import TestHereLinks
from .main import TestPhrases
from .main import TestCodeBlockOrAlertEndsSection
from .main import TestContractions
from .main import TestCodeFormatter
from .main import TestLeadingColon
from .main import rplint

__version__ = "0.5.0"
